import pytest
import os
from unittest.mock import patch


def pytest_configure(config):
    """Configure pytest"""
    os.environ["TESTING"] = "true"


@pytest.fixture
def mock_databases():
    """Mock the database initialization"""
    with patch("src.app.DATABASES", ({}, {}, {}, {})):
        with patch("src.initialization.PESTUDIO_STRINGS", {}):
            yield



from collections import Counter
import datetime
from typing import Any, List

from typing import Tuple

import os
import re
import logging


from stringzz import TokenInfo


def _get_uint_string(magic):
    print(magic)
    return f"uint16(0) == 0x{hex(magic[1])[2:]}{hex(magic[0])[2:]}"


def _sanitize_rule_name(path: str, file: str) -> str:
    """Generate a valid YARA rule name from path and filename.

    - Prefix with folder name if too short
    - Ensure it doesn't start with a number
    - Replace invalid chars with underscore
    - De-duplicate underscores
    """
    file_base = os.path.splitext(file)[0]
    cleaned = file_base
    if len(file_base) < 8:
        cleaned = path.split("\\")[-1:][0] + "_" + cleaned
    if re.search(r"^[0-9]", cleaned):
        cleaned = "sig_" + cleaned
    cleaned = re.sub(r"[^\w]", "_", cleaned)
    cleaned = re.sub(r"_+", "_", cleaned)
    return cleaned


def _get_timestamp_basic(date_obj=None):
    return (
        date_obj.strftime("%Y-%m-%d")
        if date_obj
        else datetime.datetime.now().strftime("%Y-%m-%d")
    )


def _get_file_range(size, fm_size):
    size_string = ""
    # max sample size - args.filesize_multiplier times the original size
    max_size_b = size * fm_size
    # Minimum size
    if max_size_b < 1024:
        max_size_b = 1024
    # in KB
    max_size = int(max_size_b / 1024)
    max_size_kb = max_size
    # Round
    if len(str(max_size)) == 2:
        max_size = int(round(max_size, -1))
    elif len(str(max_size)) == 3:
        max_size = int(round(max_size, -2))
    elif len(str(max_size)) == 4:
        max_size = int(round(max_size, -3))
    elif len(str(max_size)) >= 5:
        max_size = int(round(max_size, -3))
    size_string = f"filesize < {max_size}KB"
    logging.getLogger("yarobot").debug(
        "File Size Eval: SampleSize (b): %s SizeWithMultiplier (b/Kb): %s / %s RoundedSize: %s",
        str(size),
        str(max_size_b),
        str(max_size_kb),
        str(max_size),
    )
    return size_string


def _add_conditions(
    conditions,
    subconditions,
    rule_strings,
    rule_opcodes,
    high_scoring_strings,
    pe_conditions_add,
):
    # String combinations
    cond_op = ""  # opcodes condition
    cond_hs = ""  # high scoring strings condition
    cond_ls = ""  # low scoring strings condition

    low_scoring_strings = len(rule_strings) - high_scoring_strings
    if high_scoring_strings > 0:
        cond_hs = "1 of ($x*)"
    if low_scoring_strings > 0:
        if low_scoring_strings > 10:
            if high_scoring_strings > 0:
                cond_ls = "4 of them"
            else:
                cond_ls = "8 of them"
        else:
            cond_ls = "all of them"

    # If low scoring and high scoring
    cond_combined = "all of them"
    needs_brackets = False
    if low_scoring_strings > 0 and high_scoring_strings > 0:
        # If PE conditions have been added, don't be so strict with the strings
        if pe_conditions_add:
            cond_combined = "{0} or {1}".format(cond_hs, cond_ls)
            needs_brackets = True
        else:
            cond_combined = "{0} and {1}".format(cond_hs, cond_ls)
    elif low_scoring_strings > 0 and not high_scoring_strings > 0:
        cond_combined = "{0}".format(cond_ls)
    elif not low_scoring_strings > 0 and high_scoring_strings > 0:
        cond_combined = "{0}".format(cond_hs)
    if rule_opcodes:
        cond_op = " and all of ($op*)"
        # Opcodes (if needed)
    if cond_op or needs_brackets:
        subconditions.append("( {0}{1} )".format(cond_combined, cond_op))
    else:
        subconditions.append(cond_combined)


class RuleGenerator:
    def __init__(self, args, scoring_engine):
        self.prefix = args.prefix
        self.author = args.author
        self.ref = args.ref
        self.args = args
        self.scoring_engine = scoring_engine
        self.pe_module_necessary = False

    def _generate_general_condition(
        self, file_info, nofilesize, filesize_multiplier, noextras
    ):
        """
        Generates a general condition for a set of files
        :param file_info:
        :return:
        """
        conditions = []

        # Different Magic Headers and File Sizes
        magic_headers = []
        file_sizes = []
        imphashes = []

        for filePath in file_info:
            if not file_info[filePath].magic:
                continue
            magic = file_info[filePath].magic
            size = file_info[filePath].size
            imphash = file_info[filePath].imphash

            # Add them to the lists
            if magic not in magic_headers and magic != "":
                magic_headers.append(magic)
            if size not in file_sizes:
                file_sizes.append(size)
            if imphash not in imphashes and imphash != "":
                imphashes.append(imphash)

        # If different magic headers are less than 5
        if len(magic_headers) <= 5:
            magic_string = " or ".join(_get_uint_string(h) for h in magic_headers)
            if " or " in magic_string:
                conditions.append("( {0} )".format(magic_string))
            else:
                conditions.append("{0}".format(magic_string))

        # Biggest size multiplied with maxsize_multiplier
        if not nofilesize and len(file_sizes) > 0:
            conditions.append(_get_file_range(max(file_sizes), filesize_multiplier))

        # If different magic headers are less than 5
        if len(imphashes) == 1 and not noextras:
            conditions.append('pe.imphash() == "{0}"'.format(imphashes[0]))
            self.pe_module_necessary = True

        # If enough attributes were special
        condition_string = " and ".join(conditions)

        return condition_string

    def generate_rules(
        self,
        file_strings,
        file_opcodes,
        file_utf16strings,
        super_rules,
        opcode_super_rules,
        utf16_super_rules,
        file_info,
    ):
        fdata = ""
        # General Info
        general_info = "/*\n"
        general_info += "   YARA Rule Set\n"
        general_info += f"   Author: {self.args.author}\n"
        general_info += f"   Date: {_get_timestamp_basic()}\n"
        general_info += f"   Identifier: {self.args.identifier}\n"
        general_info += f"   Reference: {self.args.ref}\n"
        if license != "":
            general_info += f"   License: {self.args.license}\n"
        general_info += "*/\n\n"
        fdata += general_info

        # GLOBAL RULES ----------------------------------------------------
        if self.args.globalrule:
            condition = self._generate_general_condition(
                file_info,
                self.args.nofilesize,
                self.args.filesize_multiplier,
                self.args.noextras,
            )

            # Global Rule
            if condition != "":
                global_rule = (
                    "/* Global Rule -------------------------------------------------------------- */\n"
                    "/* Will be evaluated first, speeds up scanning process, remove at will */\n\n"
                    "global private rule gen_characteristics {\n"
                    "\tcondition:\n"
                    f"\t\t{condition}\n}}\n\n"
                )

                fdata += global_rule

        # General vars
        rules = ""
        printed_rules = {}
        rule_count = 0
        super_rule_count = 0

        # PROCESS SIMPLE RULES ----------------------------------------------------
        logging.getLogger("yarobot").info("[+] Generating Simple Rules ...")

        # logging.getLogger("yarobot").info(file_strings)

        # GENERATE SIMPLE RULES -------------------------------------------
        fdata += "/* Rule Set ----------------------------------------------------------------- */\n\n"
        all_files_set = set(file_strings.keys())
        if self.args.get_opcodes:
            all_files_set.update(file_opcodes.keys())
        all_files_set.update(file_utf16strings.keys())

        for filePath in all_files_set:
            if rule := self.generate_simple_rule(
                printed_rules,
                file_strings[filePath] if filePath in file_strings.keys() else [],
                (
                    file_opcodes[filePath]
                    if self.args.get_opcodes and filePath in file_opcodes.keys()
                    else []
                ),
                (
                    file_utf16strings[filePath]
                    if filePath in file_utf16strings.keys()
                    else []
                ),
                file_info[filePath],
                filePath,
            ):
                rules += rule
                rule_count += 1

        # GENERATE SUPER RULES --------------------------------------------
        if not self.args.nosuper:
            rules += "/* Super Rules ------------------------------------------------------------- */\n\n"
            super_rule_names = []

            print("[+] Generating Super Rules ...")
            printed_combi = {}
            for super_rule in super_rules + opcode_super_rules + utf16_super_rules:
                rules += self.generate_super_rule(
                    super_rule,
                    file_info,
                    printed_rules,
                    super_rule_names,
                    printed_combi,
                    super_rule_count,
                    None,
                )
                super_rule_count += 1

        # WRITING RULES TO FILE
        # PE Module -------------------------------------------------------
        if not self.args.noextras:
            if "pe." in rules:
                fdata += 'import "pe"\n\n'
        # RULES ------------------------------
        fdata += rules
        # Print rules to command line -------------------------------------
        logging.getLogger("yarobot").debug(rules)

        return (rule_count, super_rule_count, fdata)

    def format_rule(
        self, rule_name, file, hashes, rule_strings, rule_opcodes, conditions
    ):
        # Print rule title
        rule = (
            f"rule {rule_name} {{\n"
            f"\tmeta:\n"
            f'\t\tdescription = "{self.prefix} - file {file}"\n'
            f'\t\tauthor = "{self.author}"\n'
            f'\t\treference = "{self.ref}"\n'
            f'\t\tdate = "{_get_timestamp_basic()}"\n'
        )
        for i, hash in enumerate(hashes):
            rule += f'\t\thash{i + 1} = "{hash}"\n'
        rule += "\tstrings:\n"
        rule += "\n".join(rule_strings)
        rule += "\n"
        rule += "\n".join(rule_opcodes)
        rule += "\n\tcondition:\n"
        rule += "\t\t%s\n" % conditions
        rule += "}\n\n"
        return rule

    def generate_simple_rule(
        self,
        printed_rules,
        strings: List[TokenInfo] | None,
        opcodes: List[TokenInfo] | None,
        utf16strings: List[TokenInfo] | None,
        info,
        fname,
    ) -> str:
        if not strings and not utf16strings:
            logging.getLogger("yarobot").warning(
                "[W] Not enough high scoring strings to create a rule. (Try -z 0 to reduce the min score or --opcodes to include opcodes) FILE: %s",
                fname,
            )
            return False
        # Skip if there is nothing to do
        logging.getLogger("yarobot").info(
            "[+] Generating rule for %s, %d strings, %d opcodes, %d utf16strs",
            fname,
            len(strings),
            len(opcodes),
            len(utf16strings),
        )

        # Print rule title ----------------------------------------

        (path, file) = os.path.split(fname)
        # Prepare name via helper
        cleanedName = _sanitize_rule_name(path, file)
        # Check if already printed
        if cleanedName in printed_rules:
            printed_rules[cleanedName] += 1
            cleanedName = cleanedName + "_" + str(printed_rules[cleanedName])
        else:
            printed_rules[cleanedName] = 1

        # Condition -----------------------------------------------
        # Conditions list (will later be joined with 'or')
        conditions = []  # AND connected
        subconditions = []  # OR connected

        # Condition PE
        # Imphash and Exports - applicable to PE files only
        condition_pe = []
        condition_pe_part1 = []
        condition_pe_part2 = []

        def add_extras():
            # Add imphash - if certain conditions are met
            if (
                info.imphash not in self.scoring_engine.good_imphashes_db
                and info.imphash != ""
            ):
                # Comment to imphash
                imphash = info.imphash
                comment = ""
                # Add imphash to condition
                condition_pe_part1.append(
                    'pe.imphash() == "{0}"{1}'.format(imphash, comment)
                )
                self.pe_module_necessary = True
            if info.exports:
                e_count = 0
                for export in info.exports:
                    if export not in self.scoring_engine.good_exports_db:
                        condition_pe_part2.append('pe.exports("{0}")'.format(export))
                        e_count += 1
                        self.pe_module_necessary = True
                    if e_count > 5:
                        break

        if not self.args.noextras and info.magic.startswith(b"MZ"):
            add_extras()

        # 1st Part of Condition 1
        def add_basic_conditions(conditions, info, args):
            basic_conditions: List[Any] = []
            # Filesize
            if not args.nofilesize:
                basic_conditions.insert(
                    0, _get_file_range(info.size, args.filesize_multiplier)
                )
            # Magic
            if info.magic != b"":
                uint_string = _get_uint_string(info.magic)
                basic_conditions.insert(0, uint_string)
            conditions.append(" and ".join(basic_conditions))

        add_basic_conditions(conditions, info, self.args)
        # Add extra PE conditions to condition 1
        pe_conditions_add = False
        if condition_pe_part1 or condition_pe_part2:
            if len(condition_pe_part1) == 1:
                condition_pe.append(condition_pe_part1[0])
            elif len(condition_pe_part1) > 1:
                condition_pe.append(f"( {' or '.join(condition_pe_part1)} )")
            if len(condition_pe_part2) == 1:
                condition_pe.append(condition_pe_part2[0])
            elif len(condition_pe_part2) > 1:
                condition_pe.append(f"({' and '.join(condition_pe_part2)} )")
            # Marker that PE conditions have been added
            pe_conditions_add = True
            # Add to sub condition
            subconditions.append(" and ".join(condition_pe))

        rule_strings, high_scoring_strings, rule_opcodes = self._generate_rule_tokens(
            strings, utf16strings, opcodes, self.args.opcode_num
        )
        _add_conditions(
            conditions,
            subconditions,
            rule_strings,
            rule_opcodes,
            high_scoring_strings,
            pe_conditions_add,
        )

        # Now add string condition to the conditions
        if len(subconditions) == 1:
            conditions.append(subconditions[0])
        elif len(subconditions) > 1:
            conditions.append("( %s )" % " or ".join(subconditions))

        # Create condition string
        condition_string = "\n\t\tand ".join(conditions)

        return self.format_rule(
            cleanedName,
            file,
            [info.sha256],
            rule_strings,
            rule_opcodes,
            condition_string,
        )

    def _generate_rule_tokens(
        self, strings, utf16strings, opcodes, opcode_num, comments: bool = True
    ) -> Tuple[List[str], int, List[str]]:
        # Rule String generation
        (
            rule_strings,
            high_scoring_strings,
        ) = self.scoring_engine.generate_rule_strings(
            self.args.high_scoring,
            self.args.strings_per_rule,
            (strings or []) + (utf16strings or []),
            comments,
        )  # generate_rule_strings(args,(strings or []) + (utf16strings or []),)

        rule_opcodes = []
        if opcodes:
            rule_opcodes = _generate_rule_opcodes(opcodes, opcode_num)
        return rule_strings, high_scoring_strings, rule_opcodes

    def generate_super_rule(
        self,
        super_rule,
        infos,
        printed_rules,
        super_rule_names,
        printed_combi,
        super_rule_count,
        opcodes,
    ):
        # Prepare Name
        rule_name = ""
        file_list = []
        hashes = []
        # Loop through files
        print("Generating super rule for %s" % super_rule)
        imphashes = Counter()
        for filePath in super_rule.files:
            (path, file) = os.path.split(filePath)
            file_list.append(file)
            # Prepare name via helper
            cleanedName = _sanitize_rule_name(path, file)
            # Append it to the full name
            rule_name += "_" + cleanedName
            # Check if imphash of all files is equal
            imphash = infos[filePath].imphash
            hashes.append(infos[filePath].sha256)
            if imphash != "-" and imphash != "":
                imphashes.update([imphash])

        # Imphash usable
        if len(imphashes) == 1:
            unique_imphash = list(imphashes.items())[0][0]
            if unique_imphash in self.scoring_engine.good_imphashes_db:
                unique_imphash = ""

        # Shorten rule name
        rule_name = rule_name[:124]
        # Add count if rule name already taken
        if rule_name not in super_rule_names:
            rule_name = "%s_%s" % (rule_name, super_rule_count)
        super_rule_names.append(rule_name)

        # File name starts with a number
        if re.search(r"^[0-9]", rule_name):
            rule_name = "sig_" + rule_name
        # clean name from all characters that would cause errors
        rule_name = re.sub(r"[^\w]", "_", rule_name)
        # Check if already printed
        if rule_name in printed_rules:
            printed_combi[rule_name] += 1
            rule_name = rule_name + "_" + str(printed_combi[rule_name])
        else:
            printed_combi[rule_name] = 1

        rule_strings, high_scoring_strings, rule_opcodes = self._generate_rule_tokens(
            super_rule.strings, [], opcodes, self.args.opcode_num
        )

        # Condition -----------------------------------------------
        # Conditions list (will later be joined with 'or')
        conditions = []
        subconditions = []
        # 1st condition
        # Evaluate the general characteristics
        file_info_super = {}
        for filePath in super_rule.files:
            file_info_super[filePath] = infos[filePath]
        self._generate_general_condition(
            infos,
            self.args.nofilesize,
            self.args.filesize_multiplier,
            self.args.noextras,
        )

        # 2nd condition
        # String combinations
        _add_conditions(
            conditions,
            subconditions,
            rule_strings,
            rule_opcodes,
            high_scoring_strings,
            self.pe_module_necessary,
        )
        # Now add string condition to the conditions
        if len(subconditions) == 1:
            conditions.append(subconditions[0])
        elif len(subconditions) > 1:
            conditions.append("( %s )" % " or ".join(subconditions))
        # Create condition string
        condition_string = "\n      ) or ( ".join(conditions)

        return self.format_rule(
            rule_name,
            ", ".join(file_list),
            hashes,
            rule_strings,
            rule_opcodes,
            condition_string,
        )


def _generate_opcode_repr(i, opcode):
    return f"\t\t$op{i} = {{{opcode.reprz}}}"


def _generate_rule_opcodes(opcode_elements, opcodes_per_rule):
    # Adding the opcodes --------------------------------------
    rule_opcodes = []
    for i, opcode in enumerate(opcode_elements):
        rule_opcodes.append(_generate_opcode_repr(i, opcode))
        if i >= opcodes_per_rule:
            break
    return rule_opcodes



import os

# Resolve DB path relative to project root (two levels up from this file)
_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__)))
# DB_PATH = os.path.normpath(os.path.join(_ROOT, "dbs"))


PE_STRINGS_FILE = os.path.normpath(os.path.join(_ROOT, "3rdparty", "strings.xml"))

"../3rdparty/strings.xml"

RELEVANT_EXTENSIONS = [
    "asp",
    "vbs",
    "ps",
    "ps1",
    "tmp",
    "bas",
    "bat",
    "cmd",
    "com",
    "cpl",
    "crt",
    "dll",
    "exe",
    "msc",
    "scr",
    "sys",
    "vb",
    "vbe",
    "vbs",
    "wsc",
    "wsf",
    "wsh",
    "input",
    "war",
    "jsp",
    "php",
    "asp",
    "aspx",
    "psd1",
    "psm1",
    "py",
]



from pathlib import Path
from types import SimpleNamespace
from src.yarobot.generate import process_folder
import stringzz
import yara
import pstats

import cProfile


def test_get_pe_info_fast_rejects():
    # Not a PE
    fi = stringzz.get_file_info(b"\x7fELF......")
    assert fi.imphash == ""
    assert fi.exports == []

    # MZ but no PE signature
    fake_mz = bytearray(b"MZ" + b"\x00" * 0x3A + b"\x00\x00\x00\x00" + b"\x00" * 64)
    fi = stringzz.get_file_info(bytes(fake_mz))
    assert fi.imphash == ""
    assert fi.exports == []


def test_create_rust_struc():
    x = stringzz.TokenInfo("wasd", 16, stringzz.TokenType.BINARY, {"file", "file2"}, [""])
    print(str(x))


def test_integration(shared_datadir):
    # pr = cProfile.Profile()
    # pr.enable()

    args = SimpleNamespace(
        max_file_size=10,
        debug=False,
        max_size=128,
        min_size=4,
        get_opcodes=False,
        b="",
        recursive=True,
        oe=False,
        c=False,
        excludegood=False,
        min_score=1,
        superrule_overlap=5,
        prefix="test",
        author="test",
        ref="test",
        output_rule_file="test.yar",
        identifier="test",
        license="test",
        globalrule=True,
        nofilesize=False,
        filesize_multiplier=3,
        noextras=True,
        opcode_num=3,
        score=True,
        high_scoring=10,
        strings_per_rule=10,
        nosuper=False,
        max_file_count=10000
    )
    data = shared_datadir.joinpath("binary").read_bytes()[
        : 1024 * 1024 * args.max_file_size
    ]

    rules = process_folder(args, str(shared_datadir))
    # pr.disable()

    # stats = pstats.Stats(pr)
    # stats.sort_stats("cumulative").print_stats(10)  # Sort by cumulative time and print top 10
    r = yara.compile(source=rules)
    m = r.match(data=data)
    print(rules)
    assert len(m) > 0
    print(m)



#!/usr/bin/env python
"""
HTTP Service for yarobot - YARA Rule Generator
File upload only version
"""

from argparse import ArgumentParser
import os
import tempfile
import uuid
import click
from flask import Flask, render_template, request, jsonify
from werkzeug.utils import secure_filename
import logging
from typing import Dict, List, Optional, Any, Tuple
from .generate import process_buffers
from .common import (
    load_databases,
    initialize_pestudio_strings,
    getPrefix,
    getReference,
)
import stringzz
from dataclasses import dataclass, field, asdict


DB_PATH = os.environ.get("YAROBOT_DB_PATH", "./dbs")
MAX_FILES = 10

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("yarobot-service")

current_dir = os.path.dirname(os.path.abspath(__file__))
template_dir = os.path.join(current_dir, "templates")

app = Flask(
    __name__,
    template_folder=template_dir,
    static_folder=os.path.join(current_dir, "static"),
)

app.config["MAX_CONTENT_LENGTH"] = 10 * 1024 * 1024  # 100MB max file size

# Global variables for databases (loaded once at startup)
DATABASES = None
PESTUDIO_STRINGS = None
FP = None
SE = None


@dataclass
class DatabaseContext:
    """Container for all database-related objects"""

    good_strings_db: Optional[Any] = None
    good_opcodes_db: Optional[Any] = None
    good_imphashes_db: Optional[Any] = None
    good_exports_db: Optional[Any] = None
    pestudio_strings: Optional[Any] = None
    fp: Optional[Any] = None
    se: Optional[Any] = None

    @property
    def databases_tuple(self) -> Tuple:
        """Return databases as a tuple for compatibility with existing code"""
        return (
            self.good_strings_db,
            self.good_opcodes_db,
            self.good_imphashes_db,
            self.good_exports_db,
        )

    def is_initialized(self) -> bool:
        """Check if all databases are loaded"""
        return all(
            [
                self.good_strings_db is not None,
                self.good_opcodes_db is not None,
                self.good_imphashes_db is not None,
                self.good_exports_db is not None,
                self.pestudio_strings is not None,
            ]
        )


@dataclass
class AnalysisParameters:
    """Analysis parameters with type conversions and validation"""

    # Integer parameters with defaults
    min_size: int = 8
    min_score: int = 5
    high_scoring: int = 30
    max_size: int = 128
    strings_per_rule: int = 15
    filesize_multiplier: int = 2
    max_file_size: int = 2
    opcode_num: int = 3
    superrule_overlap: int = 5

    # Boolean parameters with defaults
    excludegood: bool = False
    score: bool = True
    nomagic: bool = False
    nofilesize: bool = False
    only_executable: bool = False
    noextras: bool = False
    debug: bool = False
    trace: bool = False
    nosimple: bool = False
    globalrule: bool = False
    nosuper: bool = False
    recursive: bool = False
    get_opcodes: bool = False

    # String parameters with defaults
    author: str = "yarobot Web Service"
    ref: str = "https://github.com/ogre2007/yarobot"
    license: str = ""
    prefix: str = "Auto-generated rule"
    identifier: str = "not set"

    def to_config(self) -> stringzz.Config:
        """Convert to stringzz.Config object"""
        return stringzz.Config(
            min_string_len=self.min_size,
            max_string_len=self.max_size,
            max_file_size_mb=self.max_size,
            extract_opcodes=self.get_opcodes,
        )

    @classmethod
    def from_form_data(cls, form_data: Dict[str, str]) -> "AnalysisParameters":
        """Create AnalysisParameters from Flask form data"""
        processed = {}

        # Integer field conversions
        int_fields = [
            "min_size",
            "min_score",
            "high_scoring",
            "max_size",
            "strings_per_rule",
            "filesize_multiplier",
            "max_file_size",
            "opcode_num",
            "superrule_overlap",
        ]

        for field in int_fields:
            if field in form_data:
                try:
                    processed[field] = int(form_data[field])
                except (ValueError, TypeError):
                    raise ValueError(f"Invalid integer value for {field}")

        # Boolean field conversions
        bool_fields = [
            "excludegood",
            "score",
            "nomagic",
            "nofilesize",
            "only_executable",
            "noextras",
            "debug",
            "trace",
            "nosimple",
            "globalrule",
            "nosuper",
            "recursive",
            "get_opcodes",
        ]

        bool_mapping = {
            "show_scores": "score",
            "no_magic": "nomagic",
            "no_filesize": "nofilesize",
            "no_extras": "noextras",
            "no_simple_rules": "nosimple",
            "global_rules": "globalrule",
            "no_super_rules": "nosuper",
        }

        for form_field, dataclass_field in bool_mapping.items():
            if form_field in form_data:
                processed[dataclass_field] = form_data[form_field].lower() in [
                    "true",
                    "1",
                    "yes",
                    "on",
                ]

        for field in bool_fields:
            if field in form_data and field not in processed:
                processed[field] = form_data[field].lower() in [
                    "true",
                    "1",
                    "yes",
                    "on",
                ]

        # String field mappings
        string_mapping = {
            "reference": "ref",
            "author": "author",
            "license": "license",
            "prefix": "prefix",
            "identifier": "identifier",
        }

        for form_field, dataclass_field in string_mapping.items():
            if form_field in form_data:
                processed[dataclass_field] = form_data[form_field]

        return cls(**processed)

    def to_dict(self) -> Dict:
        """Convert to dictionary for JSON serialization"""
        return asdict(self)


# Add form configuration constants
FORM_FIELDS_CONFIG = {
    "basic": [
        {
            "name": "min_score",
            "type": "number",
            "label": "Minimum Score",
            "default": 5,
            "min": 0,
            "max": 100,
        },
        {
            "name": "author",
            "type": "text",
            "label": "Author",
            "default": "yarobot Web Interface",
        },
        {
            "name": "get_opcodes",
            "type": "checkbox",
            "label": "Use Opcodes",
            "form_field": "use-opcodes",
            "default": False,
        },
        {
            "name": "globalrule",
            "type": "checkbox",
            "label": "Generate Global Rules",
            "form_field": "global_rules",
            "default": False,
        },
        {
            "name": "excludegood",
            "type": "checkbox",
            "label": "Exclude good",
            "form_field": "excludegood",
            "default": False,
        },
    ],
    "advanced": [
        {
            "name": "min_size",
            "type": "number",
            "label": "Minimum String Size",
            "default": 8,
            "min": 1,
        },
        {
            "name": "max_size",
            "type": "number",
            "label": "Maximum String Size",
            "default": 128,
            "min": 1,
        },
        {
            "name": "strings_per_rule",
            "type": "number",
            "label": "Strings per Rule",
            "default": 15,
            "min": 1,
        },
        {
            "name": "high_scoring",
            "type": "number",
            "label": "High Scoring Threshold",
            "default": 30,
            "min": 0,
        },
        {
            "name": "nomagic",
            "type": "checkbox",
            "label": "No Magic",
            "form_field": "no_magic",
            "default": False,
        },
        {
            "name": "nofilesize",
            "type": "checkbox",
            "label": "No Filesize",
            "form_field": "no_filesize",
            "default": False,
        },
        {
            "name": "noextras",
            "type": "checkbox",
            "label": "No Extras",
            "form_field": "no_extras",
            "default": False,
        },
        {
            "name": "nosimple",
            "type": "checkbox",
            "label": "No Simple Rules",
            "form_field": "no_simple_rules",
            "default": False,
        },
        {
            "name": "nosuper",
            "type": "checkbox",
            "label": "No Super Rules",
            "form_field": "no_super_rules",
            "default": False,
        },
        {
            "name": "recursive",
            "type": "checkbox",
            "label": "Recursive Analysis",
            "default": False,
        },
        {
            "name": "prefix",
            "type": "text",
            "label": "Rule Prefix",
            "default": "Auto-generated rule",
        },
        {"name": "license", "type": "text", "label": "License", "default": ""},
        {
            "name": "opcode_num",
            "type": "number",
            "label": "Opcode Number",
            "default": 3,
            "min": 1,
        },
    ],
}


@dataclass
class AnalysisRequest:
    """Complete analysis request with parameters and files"""

    parameters: AnalysisParameters
    files: List[Any]
    identifier: str = field(default_factory=lambda: f"upload_{uuid.uuid4().hex[:8]}")

    def __post_init__(self):
        """Post-initialization processing"""
        # Set identifier if not already set
        if self.parameters.identifier == "not set":
            self.parameters.identifier = self.identifier
        else:
            self.identifier = self.parameters.identifier

        # Process ref and prefix
        self.parameters.ref = getReference(self.parameters.ref)
        self.parameters.prefix = getPrefix(self.parameters.prefix, self.identifier)

    @classmethod
    def from_flask_request(cls, flask_request) -> "AnalysisRequest":
        """Create AnalysisRequest from Flask request object"""
        # Get files
        if "files" not in flask_request.files:
            raise ValueError("No files provided")

        files = flask_request.files.getlist("files")
        if len(files) > MAX_FILES:
            raise ValueError(f"Too many files, max {MAX_FILES}")

        if not files or all(file.filename == "" for file in files):
            raise ValueError("No files selected")

        # Get parameters from form data
        params = AnalysisParameters.from_form_data(flask_request.form.to_dict())

        return cls(parameters=params, files=files)

    @property
    def files_count(self) -> int:
        return len(self.files)


@dataclass
class AnalysisResult:
    """Result of analysis operation"""

    status: str
    rules_content: Optional[str] = None
    error: Optional[str] = None
    rules_count: int = 0
    files_analyzed: int = 0
    identifier: str = ""

    def to_dict(self) -> Dict:
        """Convert to dictionary for JSON response"""
        result = {
            "status": self.status,
            "rules_count": self.rules_count,
            "files_analyzed": self.files_analyzed,
            "identifier": self.identifier,
        }

        if self.rules_content:
            result.update(
                {"rules_generated": True, "rules_content": self.rules_content}
            )

        if self.error:
            result["error"] = str(self.error)

        return result

    @classmethod
    def success(
        cls, rules_content: str, identifier: str, files_analyzed: int
    ) -> "AnalysisResult":
        """Create successful result"""
        return cls(
            status="success",
            rules_content=rules_content,
            rules_count=rules_content.count("rule ") if rules_content else 0,
            files_analyzed=files_analyzed,
            identifier=identifier,
        )

    @classmethod
    def error(cls, error_message: str) -> "AnalysisResult":
        """Create error result"""
        return cls(status="error", error=error_message)


# Global database context
db_context = DatabaseContext()


def initialize_databases(db_path: str):
    """Initialize databases on startup"""
    try:
        logger.info("Initializing databases...")

        # Load pestudio strings
        db_context.pestudio_strings = initialize_pestudio_strings()

        # Load databases
        (
            db_context.good_strings_db,
            db_context.good_opcodes_db,
            db_context.good_imphashes_db,
            db_context.good_exports_db,
        ) = load_databases(db_path)

        logger.info("Databases initialized successfully")

    except Exception as e:
        logger.error(f"Failed to initialize databases: {e}")
        raise


def init_analysis_context():
    """Initialize analysis context (FP, SE)"""
    config = stringzz.Config()
    global db_context
    db_context.fp, db_context.se = stringzz.init_analysis(
        config,
        False,
        5,
        5,
        *db_context.databases_tuple,
        db_context.pestudio_strings,
    )
    return db_context.fp, db_context.se


def ensure_databases_initialized():
    """Ensure databases are initialized before analysis"""
    if not db_context.is_initialized():
        initialize_databases(DB_PATH)
        init_analysis_context()


# Web Routes
@app.route("/")
def index():
    """Main web interface"""
    return render_template("index.html", form_config=FORM_FIELDS_CONFIG)


@app.route("/health", methods=["GET"])
def health_check():
    """Health check endpoint"""
    return jsonify(
        {
            "status": "healthy",
            "service": "yarobot-http",
            "databases_loaded": db_context.is_initialized(),
        }
    )


@app.route("/api/analyze", methods=["POST"])
def analyze():
    """Analyze uploaded files and generate YARA rules"""
    try:
        # Create analysis request from Flask request
        analysis_request = AnalysisRequest.from_flask_request(request)
        print(analysis_request)
        # Ensure databases are initialized
        ensure_databases_initialized()
        global db_context
        logger.info(f"Starting analysis for {analysis_request.files_count} files")
        db_context.fp = stringzz.FileProcessor(analysis_request.parameters.to_config())
        db_context.se.min_score = analysis_request.parameters.min_score
        db_context.se.excludegood = analysis_request.parameters.excludegood
        db_context.se.superrule_overlap = analysis_request.parameters.superrule_overlap
        rules_content = process_buffers(
            db_context.fp,
            db_context.se,
            analysis_request.parameters,
            [f.read() for f in analysis_request.files],
            *db_context.databases_tuple,
            db_context.pestudio_strings,
        )

        result = AnalysisResult.success(
            rules_content=rules_content,
            identifier=analysis_request.identifier,
            files_analyzed=analysis_request.files_count,
        )
        result = result.to_dict()
        # print(result)
        init_analysis_context()
        return jsonify(result)

    except ValueError as e:
        logger.error(f"Validation error: {e}")
        return jsonify(AnalysisResult.error(str(e)).to_dict()), 400
    # except Exception as e:
    #    logger.error(f"Error during analysis: {e}")
    #    return jsonify(AnalysisResult.error("Internal server error").to_dict()), 500


@app.route("/api/status", methods=["GET"])
def get_status():
    """Get service status and database information"""
    db_info = {}

    if db_context.is_initialized():
        db_info = {
            "good_strings_entries": len(db_context.good_strings_db),
            "good_opcodes_entries": len(db_context.good_opcodes_db),
            "good_imphashes_entries": len(db_context.good_imphashes_db),
            "good_exports_entries": len(db_context.good_exports_db),
            "pestudio_strings_loaded": db_context.pestudio_strings is not None,
        }

    return jsonify(
        {
            "status": "running",
            "databases": db_info,
            "databases_initialized": db_context.is_initialized(),
        }
    )


# Error handlers
@app.errorhandler(413)
def too_large(e):
    return jsonify(AnalysisResult.error("File too large").to_dict()), 413


@app.errorhandler(500)
def internal_error(e):
    return jsonify(AnalysisResult.error("Internal server error").to_dict()), 500


@app.errorhandler(404)
def not_found(e):
    return render_template("index.html", form_config=FORM_FIELDS_CONFIG), 404


def create_template_directories():
    """Create necessary directories for templates"""
    os.makedirs(template_dir, exist_ok=True)

    # Create static directory if needed
    static_dir = os.path.join(current_dir, "static")
    os.makedirs(static_dir, exist_ok=True)


@click.command()
@click.option(
    "-g", type=click.Path(exists=True), help="path to folder with goodware dbs"
)
def main(g=None):
    """Main entry point"""
    global DB_PATH

    if g:
        DB_PATH = g

    # Initialize databases before starting the server
    initialize_databases(DB_PATH)
    init_analysis_context()

    # Create templates directory if it doesn't exist
    create_template_directories()

    logger.info(
        f"Starting yarobot web interface on "
        f"http://{os.getenv('YAROBOT_HOST', '0.0.0.0')}:{os.getenv('YAROBOT_PORT', 5000)}"
    )
    logger.info(f"Template directory: {template_dir}")

    # Start Flask app
    app.run(
        host=os.getenv("YAROBOT_HOST", "0.0.0.0"),
        port=int(os.getenv("YAROBOT_PORT", 5000)),
        debug=os.getenv("YAROBOT_DEBUG", "false").lower() == "true",
    )


if __name__ == "__main__":
    main()



"""
yarGen - Yara Rule Generator, Copyright (c) 2015, Florian Roth
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the copyright owner nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL Florian Roth BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from collections import Counter
import logging
import os

import click
import stringzz
from .common import load, save
from .config import RELEVANT_EXTENSIONS

DB_PATH = "dbs"


def process_goodware_folder(
    goodware_path,
    extensions=None,
    recursive=None,
    minssize=None,
    maxssize=None,
    fsize=None,
    get_opcodes=None,
    debug=None,
    max_file_count=None,
):
    config = stringzz.Config(
        extensions=extensions,
        recursive=recursive,
        min_string_len=minssize,
        max_string_len=maxssize,
        max_file_size_mb=fsize,
        debug=debug,
        max_file_count=max_file_count,
    )
    fp = stringzz.FileProcessor(config)
    results = fp.parse_sample_dir(goodware_path)
    # print(file_infos)
    good_json = Counter({k: v.count for k, v in results.strings.items()})
    good_opcodes_json = {k: v.count for k, v in results.opcodes.items()}
    good_json.update({k: v.count for k, v in results.utf16strings.items()})
    good_exports_json = [
        fi.exports for _, fi in results.file_infos.items() if fi.exports
    ]
    good_imphashes_json = [
        fi.imphash for _, fi in results.file_infos.items() if fi.imphash
    ]

    # print(sorted(good_json.items(), key=lambda x: x[1], reverse=True)[:10])

    # Save
    save(DB_PATH, good_json, "good-strings.json")
    save(DB_PATH, good_opcodes_json, "good-opcodes.json")
    save(DB_PATH, good_exports_json, "good-exports.json")
    save(DB_PATH, good_imphashes_json, "good-imphashes.json")

    click.echo(
        "New database with %d string, %d opcode, %d imphash, %d export entries created. "
        # "(remember to use --opcodes to extract opcodes from the samples and create the opcode databases)"
        % (
            len(good_json),
            len(good_opcodes_json),
            len(good_imphashes_json),
            len(good_exports_json),
        )
    )


@click.group()
def cli():
    """database manager"""
    pass


@cli.command()
@click.argument("goodware_path", type=click.Path(exists=True), required=True)
@click.option(
    "-i", "--identifier", help="Identifier for the database files", required=True
)
@click.option(
    "--update",
    help="Update existing database with new goodware samples",
    is_flag=True,
    default=False,
)
@click.option(
    "--max-file-count", help="Max number of files to process", type=int, default=10000
)
@click.option("--debug", help="Debug output", is_flag=True, default=False)
def update(goodware_path, **kwargs):
    """Manage goodware databases"""
    args = type("Args", (), kwargs)()
    print("[+] Processing goodware files ...")
    good_strings_db, good_opcodes_db, good_imphashes_db, good_exports_db = (
        process_goodware_folder(goodware_path, max_file_count=args.max_file_count)
    )

    # Update existing databases
    if args.update:
        print("[+] Updating databases ...")

        # Evaluate the database identifiers
        db_identifier = ""
        if args.i != "":
            db_identifier = "-%s" % args.i
        strings_db = "./dbs/good-strings%s.db" % db_identifier
        opcodes_db = "./dbs/good-opcodes%s.db" % db_identifier
        imphashes_db = "./dbs/good-imphashes%s.db" % db_identifier
        exports_db = "./dbs/good-exports%s.db" % db_identifier

        # Strings -----------------------------------------------------
        print("[+] Updating %s ..." % strings_db)
        good_pickle = load(strings_db)
        print("Old string database entries: %s" % len(good_pickle))
        good_pickle.update(good_strings_db)
        print("New string database entries: %s" % len(good_pickle))
        save(good_pickle, strings_db)

        # Opcodes -----------------------------------------------------
        print("[+] Updating %s ..." % opcodes_db)
        good_opcode_pickle = load(opcodes_db)
        print("Old opcode database entries: %s" % len(good_opcode_pickle))
        good_opcode_pickle.update(good_opcodes_db)
        print("New opcode database entries: %s" % len(good_opcode_pickle))
        save(good_opcode_pickle, opcodes_db)

        # Imphashes ---------------------------------------------------
        print("[+] Updating %s ..." % imphashes_db)
        good_imphashes_pickle = load(imphashes_db)
        print("Old opcode database entries: %s" % len(good_imphashes_pickle))
        good_imphashes_pickle.update(good_imphashes_db)
        print("New opcode database entries: %s" % len(good_imphashes_pickle))
        save(good_imphashes_pickle, imphashes_db)

        # Exports -----------------------------------------------------
        print("[+] Updating %s ..." % exports_db)
        good_exports_pickle = load(exports_db)
        print("Old opcode database entries: %s" % len(good_exports_pickle))
        good_exports_pickle.update(good_exports_db)
        print("New opcode database entries: %s" % len(good_exports_pickle))
        save(good_exports_pickle, exports_db)

    if args.update:
        click.echo(f"[+] Updating goodware database with samples from {goodware_path}")
        # from app.dbs import update_goodware_db
        # update_goodware_db(args)


@cli.command()
@click.option(
    "-y",
    "--min-size",
    help="Minimum string length to consider (default=8)",
    type=int,
    default=8,
)
@click.option(
    "-s",
    "--max-size",
    help="Maximum length to consider (default=128)",
    type=int,
    default=128,
)
@click.option(
    "-R",
    "--recursive",
    help="Recursively scan directories",
    is_flag=True,
    default=False,
)
@click.option(
    "--oe",
    "--only-executable",
    help="Only scan executable extensions EXE, DLL, ASP, JSP, PHP, BIN, INFECTED",
    is_flag=True,
    default=False,
)
@click.option(
    "-fs",
    "--max-file-size",
    help="Max file size in MB to analyze (default=2)",
    type=int,
    default=5,
)
@click.option("--debug", help="Debug output", is_flag=True, default=False)
@click.option(
    "--opcodes",
    help="Do use the OpCode feature (use this if not enough high scoring strings can be found)",
    is_flag=True,
    default=False,
)
@click.option(
    "--max-file-count",
    help="Max number of files to process (default=10_000)",
    type=int,
    default=10_000,
)
@click.argument("goodware_path", type=click.Path(exists=True), required=True)
def create(goodware_path, **kwargs):
    args = type("Args", (), kwargs)()
    click.echo("[+] Creating local database ...")
    process_goodware_folder(
        goodware_path,
        RELEVANT_EXTENSIONS if args.oe else None,
        args.recursive,
        args.min_size,
        args.max_size,
        args.max_file_size,
        False,
        args.debug,
        args.max_file_count,
    )


if __name__ == "__main__":
    logging.basicConfig(level=os.environ.get("YAROBOT_LOG_LEVEL", "INFO"))
    logging.getLogger().setLevel(logging.DEBUG)
    cli()
    # Identifier



from flask import render_template
import pytest
import tempfile
import os
import json
from unittest.mock import patch, MagicMock
import sys
import io
from werkzeug.datastructures import FileStorage
import io

# Import the app module
from src.yarobot.app import (
    FORM_FIELDS_CONFIG,
    AnalysisParameters,
    app,
    AnalysisRequest,
)


@pytest.fixture
def client():
    """Create a test client for the Flask app"""
    app.config["TESTING"] = True
    app.config["WTF_CSRF_ENABLED"] = False
    with app.test_client() as client:
        yield client


@pytest.fixture
def mock_databases():
    """Mock the database initialization"""
    with patch("src.yarobot.app.DATABASES", ({}, {}, {}, {})):
        with patch("src.yarobot.app.PESTUDIO_STRINGS", {}):
            yield


class TestAnalysisRequest:
    """Test the AnalysisRequest class"""

    @pytest.mark.unit
    def test_default_parameters(self):
        """Test that default parameters are set correctly"""
        args = AnalysisParameters()

        assert args.min_size == 8
        # assert args.min_score == 5
        assert args.high_scoring == 30
        assert args.max_size == 128
        assert args.strings_per_rule == 15
        assert args.excludegood is False
        assert args.author == "yarobot Web Service"


class TestHealthEndpoints:
    """Test health and status endpoints"""

    @pytest.mark.http
    def test_health_check(self, client, mock_databases):
        """Test health check endpoint"""
        response = client.get("/health")
        assert response.status_code == 200
        data = json.loads(response.data)
        assert data["status"] == "healthy"

    @pytest.mark.http
    def test_status_endpoint(self, client, mock_databases):
        """Test status endpoint"""
        response = client.get("/api/status")
        assert response.status_code == 200
        data = json.loads(response.data)
        assert data["status"] == "running"


class TestAnalyzeEndpoint:
    """Test the main analyze endpoint"""

    @pytest.fixture
    def sample_file(self):
        """Create a sample test file"""
        with tempfile.NamedTemporaryFile(delete=False, suffix=".exe") as f:
            f.write(
                b"MZ\x90\x00\x03\x00\x00\x00\x04\x00\x00\x00\xff\xffaaaaaaaaaaaaaaaaaaaaa"
                * 35
            )
            f.flush()
            yield f.name
        if os.path.exists(f.name):
            os.unlink(f.name)

    @pytest.mark.integration
    def test_single_file_analysis(self, sample_file, client):
        """Test analysis with single file"""
        # Mock the process_folder to return sample rules
        # mock_process.return_value = "rule TestRule { condition: true }"
        # initialize_databases()

        a = io.BytesIO(b"BBBBBBBBBBBBBBBBBBBBBB\0xxxxxxxxxxxxxxxxxx")
        with open(sample_file, "rb") as f:
            # print(f.read())
            file_storage = FileStorage(
                stream=f, filename="test_file.txt", content_type="text/plain"
            )
            file_storage2 = FileStorage(
                stream=a, filename="test_file.exe", content_type="text/plain"
            )
            response = client.post(
                "/api/analyze",
                data={
                    "files": [file_storage, file_storage2],
                    "min_score": "0",
                    "get_opcodes": "true",
                },
            )

        # This should be a real Flask response, not a mock
        assert response.status_code == 200
        data = json.loads(response.data)
        assert data["status"] == "success"
        assert data["rules_generated"] is True
        assert data["rules_count"] == 2
        print(data["rules_content"])

        # Verify the mock was called
        # mock_process.assert_called_once()

    @pytest.mark.http
    def test_no_files_provided(self, client, mock_databases):
        """Test request with no files"""
        response = client.post("/api/analyze")
        assert response.status_code == 400
        data = json.loads(response.data)
        assert data["error"] == "No files provided"

    @pytest.mark.http
    def test_empty_files_list(self, client, mock_databases):
        """Test request with empty files list"""
        # Create empty files list
        response = client.post("/api/analyze", data={"files": ""})
        assert response.status_code == 400
        data = json.loads(response.data)
        assert data["error"] == "No files provided"

    @pytest.mark.http
    @patch("src.yarobot.generate.process_folder")
    def test_invalid_parameter_types(
        self, mock_process, client, mock_databases, sample_file
    ):
        """Test analysis with invalid parameter types"""
        with open(sample_file, "rb") as f:
            response = client.post(
                "/api/analyze",
                data={"files": (f, "test.exe"), "min_score": "invalid_number"},
            )

        assert response.status_code == 400
        data = json.loads(response.data)
        assert "error" in data


class TestErrorHandlers:
    """Test error handlers"""

    @pytest.mark.http
    def test_method_not_allowed(self, client, mock_databases):
        """Test 405 error handling"""
        response = client.put("/api/analyze")
        assert response.status_code == 405



#!/usr/bin/env python


"""
yarGen - Yara Rule Generator, Copyright (c) 2015, Florian Roth
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the copyright owner nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL Florian Roth BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import os
import logging

from .common import (
    getIdentifier,
    getPrefix,
    getReference,
    initialize_pestudio_strings,
    load_databases,
    common_multi_analysis_options,
    common_single_analysis_options,
)

import pstats

import cProfile
from .rule_generator import RuleGenerator

from .config import RELEVANT_EXTENSIONS

import stringzz

import click


def print_generated_stats(args, rule_count, super_rule_count):
    print("[=] Generated %s SIMPLE rules." % str(rule_count))
    if not args.nosuper:
        print("[=] Generated %s SUPER rules." % str(super_rule_count))
    if "output_rule_file" in args.__dict__.keys() and args.output_rule_file:
        print("[=] All rules written to %s" % args.output_rule_file)


def process_buffers(
    fp: stringzz.FileProcessor,
    se: stringzz.ScoringEngine,
    args,
    data: list[bytes],
    good_strings_db={},
    good_opcodes_db={},
    good_imphashes_db={},
    good_exports_db={},
    pestudio_strings={},
):
    logging.getLogger("yarobot").info(
        f"[+] Generating YARA rules from  {len(data)} buffers "
    )
    # print(fp, se)

    print("excludegood", args.excludegood)
    (
        string_combis,
        string_superrules,
        utf16_combis,
        utf16_superrules,
        opcode_combis,
        opcode_superrules,
        file_strings,
        file_opcodes,
        file_utf16strings,
        file_infos,
    ) = stringzz.analyze_buffers_comprehensive(data, fp, se)
    # print(file_strings)
    # Create Rule Files
    rg = RuleGenerator(args, se)
    print("fs", file_strings)
    (rule_count, super_rule_count, rules) = rg.generate_rules(
        file_strings,
        file_opcodes,
        file_utf16strings,
        string_superrules,
        opcode_superrules,
        utf16_superrules,
        file_infos,
    )

    print_generated_stats(args, rule_count, super_rule_count)

    return rules


def process_folder(
    args,
    folder,
    good_strings_db={},
    good_opcodes_db={},
    good_imphashes_db={},
    good_exports_db={},
    pestudio_strings={},
):
    if args.get_opcodes and len(good_opcodes_db) < 1:
        logging.getLogger("yarobot").warning("Missing goodware opcode databases.")
        args.get_opcodes = False

    if len(good_exports_db) < 1 and len(good_imphashes_db) < 1:
        logging.getLogger("yarobot").warning(
            "Missing goodware imphash/export databases."
        )

    if len(good_strings_db) < 1:
        logging.getLogger("yarobot").warning("no goodware databases found. ")

    # Scan malware files
    config = stringzz.Config(
        recursive=args.recursive,
        extensions=RELEVANT_EXTENSIONS,
        min_string_len=args.min_size,
        max_string_len=args.max_size,
        max_file_size_mb=args.max_file_size,
        extract_opcodes=args.get_opcodes,
        debug=args.debug,
        max_file_count=args.max_file_count if args.max_file_count else None,
    )
    fp, se = stringzz.init_analysis(
        config,
        args.excludegood,
        args.min_score,
        args.superrule_overlap,
        good_strings_db,
        good_opcodes_db,
        good_imphashes_db,
        good_exports_db,
        pestudio_strings,
    )

    logging.getLogger("yarobot").info(f"[+] Generating YARA rules from {folder}")
    (
        combinations,
        super_rules,
        utf16_combinations,
        utf16_super_rules,
        opcode_combinations,
        opcode_super_rules,
        file_strings,
        file_opcodes,
        file_utf16strings,
        file_info,
    ) = stringzz.process_malware(folder, fp, se)
    # Apply intelligent filters
    logging.getLogger("yarobot").info(
        "[-] Applying intelligent filters to string findings ..."
    )

    # Create Rule Files
    rg = RuleGenerator(args, se)
    (rule_count, super_rule_count, rules) = rg.generate_rules(
        file_strings,
        file_opcodes,
        file_utf16strings,
        super_rules,
        opcode_super_rules,
        utf16_super_rules,
        file_info,
    )

    print_generated_stats(args, rule_count, super_rule_count)
    return rules


@click.group()
def cli():
    pass


@cli.command()
@click.argument("malware_path", type=click.Path(exists=True))
@common_single_analysis_options
@common_multi_analysis_options
def generate(malware_path, **kwargs):
    pr = cProfile.Profile()
    pr.enable()
    """Generate YARA rules from malware samples"""
    args = type("Args", (), kwargs)()
    args.goodware_dbs
    args.identifier = getIdentifier(args.identifier, malware_path)
    print("[+] Using identifier '%s'" % args.identifier)

    # Reference
    args.ref = getReference(args.ref)
    print("[+] Using reference '%s'" % args.ref)

    # Prefix
    args.prefix = getPrefix(args.prefix, args.identifier)
    print("[+] Using prefix '%s'" % args.prefix)

    pestudio_strings = initialize_pestudio_strings()
    print("[+] Reading goodware strings from database 'good-strings.db' ...")
    print(
        "    (This could take some time and uses several Gigabytes of RAM depending on your db size)"
    )

    if args.goodware_dbs:
        good_strings_db, good_opcodes_db, good_imphashes_db, good_exports_db = (
            load_databases(args.goodware_dbs)
        )
    else:
        logging.getLogger("yarobot").warning(
            "No goodware databases found. Please create new databases."
        )
        good_strings_db, good_opcodes_db, good_imphashes_db, good_exports_db = (
            {},
            {},
            {},
            {},
        )
    # exit()
    rules = process_folder(
        args,
        malware_path,
        good_strings_db,
        good_opcodes_db,
        good_imphashes_db,
        good_exports_db,
        pestudio_strings,
    )
    with open(args.output_rule_file, "wt") as f:
        f.write(rules)
    pr.disable()

    stats = pstats.Stats(pr)
    stats.sort_stats("cumulative").print_stats(
        10
    )  # Sort by cumulative time and print top 10


# MAIN ################################################################
if __name__ == "__main__":
    logging.basicConfig(level=os.environ.get("YAROBOT_LOG_LEVEL", "INFO"))
    logging.getLogger().setLevel(logging.DEBUG)
    generate()
    # Identifier



from collections import Counter
import functools
import gzip
import logging
import os

import click
import orjson as json

from .config import PE_STRINGS_FILE

from lxml import etree


def getPrefix(prefix, identifier):
    """
    Get a prefix string for the rule description based on the identifier
    :param prefix:
    :param identifier:
    :return:
    """
    if prefix == "Auto-generated rule":
        return identifier
    else:
        return prefix


def getIdentifier(id, path):
    """
    Get a identifier string - if the provided string is the path to a text file, then read the contents and return it as
    reference, otherwise use the last element of the full path
    :param ref:
    :return:
    """
    # Identifier
    if id == "not set" or not os.path.exists(id):
        # Identifier is the highest folder name
        return os.path.basename(path.rstrip("/"))
    else:
        # Read identifier from file
        identifier = open(id).read()
        print("[+] Read identifier from file %s > %s" % (id, identifier))
        return identifier


def getReference(ref):
    """
    Get a reference string - if the provided string is the path to a text file, then read the contents and return it as
    reference
    :param ref:
    :return:
    """
    if os.path.exists(ref):
        reference = open(ref).read()
        print("[+] Read reference from file %s > %s" % (ref, reference))
        return reference
    else:
        return ref


def emptyFolder(dir):
    """
    Removes all files from a given folder
    :return:
    """
    for file in os.listdir(dir):
        filePath = os.path.join(dir, file)
        try:
            if os.path.isfile(filePath):
                print("[!] Removing %s ..." % filePath)
                os.unlink(filePath)
        except Exception as e:
            print(e)


def initialize_pestudio_strings():
    # if not os.path.isfile(get_abs_path(PE_STRINGS_FILE)):
    #    return None
    print("[+] Processing PEStudio strings ...")

    pestudio_strings = {}

    tree = etree.parse(PE_STRINGS_FILE)
    processed_strings = {}
    pestudio_strings["strings"] = tree.findall(".//string")
    pestudio_strings["av"] = tree.findall(".//av")
    pestudio_strings["folder"] = tree.findall(".//folder")
    pestudio_strings["os"] = tree.findall(".//os")
    pestudio_strings["reg"] = tree.findall(".//reg")
    pestudio_strings["guid"] = tree.findall(".//guid")
    pestudio_strings["ssdl"] = tree.findall(".//ssdl")
    pestudio_strings["ext"] = tree.findall(".//ext")
    pestudio_strings["agent"] = tree.findall(".//agent")
    pestudio_strings["oid"] = tree.findall(".//oid")
    pestudio_strings["priv"] = tree.findall(".//priv")
    for category, elements in pestudio_strings.items():
        for elem in elements:
            processed_strings[elem.text] = (5, category)
    return processed_strings


def load_databases(db_path):
    if not db_path or not os.path.isdir(db_path):
        logging.getLogger("yarobot").error("Database directory not found %s", db_path)
        return ({}, {}, {}, {})
    good_strings_db = Counter()
    good_opcodes_db = Counter()
    good_imphashes_db = Counter()
    good_exports_db = Counter()

    # Initialize all databases
    for file in os.listdir(db_path):
        if file.endswith(".db") or file.endswith(".json"):
            if file.startswith("good-strings"):
                load_db(
                    db_path,
                    file,
                    good_strings_db,
                    True if file.endswith(".json") else False,
                )
            if file.startswith("good-opcodes"):
                load_db(
                    db_path,
                    file,
                    good_opcodes_db,
                    True if file.endswith(".json") else False,
                )
            if file.startswith("good-imphashes"):
                pass  # load_db(file, good_imphashes_db, True if file.endswith(".json") else False) TODO
            if file.startswith("good-exports"):
                pass  # load_db(file, good_exports_db, True if file.endswith(".json") else False) TODO
    return good_strings_db, good_opcodes_db, good_imphashes_db, good_exports_db


def common_single_analysis_options(f):
    @click.option(
        "-g",
        "--goodware-dbs",
        help="Goodware databases",
        type=click.Path(exists=True),
    )
    @click.option(
        "-y",
        "--min-size",
        help="Minimum string length to consider (default=8)",
        type=int,
        default=8,
    )
    @click.option(
        "-z",
        "--min-score",
        help="Minimum score to consider (default=5)",
        type=int,
        default=5,
    )
    @click.option(
        "-x",
        "--high-scoring",
        help='Score required to set string as "highly specific string" (default: 30)',
        type=int,
        default=30,
    )
    @click.option(
        "-s",
        "--max-size",
        help="Maximum length to consider (default=128)",
        type=int,
        default=128,
    )
    @click.option(
        "-rc",
        "--strings-per-rule",
        help="Maximum number of strings per rule (default=15, intelligent filtering will be applied)",
        type=int,
        default=15,
    )
    @click.option(
        "--excludegood",
        help="Force the exclude all goodware strings",
        is_flag=True,
        default=False,
    )
    @click.option(
        "-o", "--output-rule-file", help="Output rule file", default="yarobot_rules.yar"
    )
    @click.option(
        "-e",
        "--output-dir-strings",
        help="Output directory for string exports",
        default="",
    )
    @click.option(
        "-a", "--author", help="Author Name", default="yarobot Rule Generator"
    )
    @click.option(
        "--ref",
        help="Reference (can be string or text file)",
        default="https://github.com/ogre2007/yarobot",
    )
    @click.option("-l", "--license", help="License", default="")
    @click.option(
        "-p",
        "--prefix",
        help="Prefix for the rule description",
        default="Auto-generated rule",
    )
    @click.option(
        "-b",
        "--identifier",
        help="Text file from which the identifier is read (default: last folder name in the full path)",
        default="not set",
    )
    @click.option(
        "--score",
        help="Show the string scores as comments in the rules",
        is_flag=True,
        default=False,
    )
    @click.option(
        "--nomagic",
        help="Don't include the magic header condition statement",
        is_flag=True,
        default=False,
    )
    @click.option(
        "--nofilesize",
        help="Don't include the filesize condition statement",
        is_flag=True,
        default=False,
    )
    @click.option(
        "-fm",
        "--filesize-multiplier",
        help="Multiplier for the maximum 'filesize' condition value (default: 2)",
        type=int,
        default=2,
    )
    @click.option(
        "--oe",
        "--only-executable",
        help="Only scan executable extensions EXE, DLL, ASP, JSP, PHP, BIN, INFECTED",
        is_flag=True,
        default=False,
    )
    @click.option(
        "-fs",
        "--max-file-size",
        help="Max file size in MB to analyze (default=2)",
        type=int,
        default=2,
    )
    @click.option(
        "--noextras",
        help="Don't use extras like Imphash or PE header specifics",
        is_flag=True,
        default=False,
    )
    @click.option("--debug", help="Debug output", is_flag=True, default=False)
    @click.option("--trace", help="Trace output", is_flag=True, default=False)
    @click.option(
        "--get-opcodes",
        help="Do use the OpCode feature (use this if not enough high scoring strings can be found)",
        is_flag=True,
        default=False,
    )
    @click.option(
        "-n",
        "--opcode-num",
        help="Number of opcodes to add if not enough high scoring string could be found (default=3)",
        type=int,
        default=3,
    )
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper


def common_multi_analysis_options(f):
    @click.option(
        "-w",
        "--superrule-overlap",
        help="Minimum number of strings that overlap to create a super rule (default: 5)",
        type=int,
        default=5,
    )
    @click.option(
        "--nosimple",
        help="Skip simple rule creation for files included in super rules",
        is_flag=True,
        default=False,
    )
    @click.option(
        "--globalrule",
        help="Create global rules (improved rule set speed)",
        is_flag=True,
        default=False,
    )
    @click.option(
        "--nosuper",
        help="Don't try to create super rules that match against various files",
        is_flag=True,
        default=False,
    )
    @click.option(
        "-R",
        "--recursive",
        help="Recursively scan directories",
        is_flag=True,
        default=False,
    )
    @click.option(
        "--max-file-count",
        help="Max number of files to process",
        type=int,
        default=10000,
    )
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper


def load(filename, just_json=False):
    if just_json:
        with open(filename, "rb") as file:
            return json.loads(file.read())
    else:
        file = gzip.GzipFile(filename, "rb")
        object = json.loads(file.read())
        file.close()
        return object


def save(db_path, object, filename):
    os.makedirs(db_path, exist_ok=True)
    path = os.path.join(db_path, filename)
    with open(path, "wb") as file:
        file.write(bytes(json.dumps(object)))


def load_db(db_path, file, local_counter, just_json=False):
    filePath = os.path.join(db_path, file)
    print("[+] Loading %s ..." % filePath)
    before = len(local_counter)
    js = load(filePath, just_json)
    local_counter.update(js)
    added = len(local_counter) - before
    print("[+] Total: %s / Added %d entries" % (len(local_counter), added))

    return len(local_counter)
